"""User-related API routes."""

import ipaddress
import logging
from typing import Annotated

from ab_client.auth_client import LoginRequest, OAuth2TokenExposed, RefreshTokenRequest
from ab_core.identity_context.dependency import IdentityContext, get_identity_context
from fastapi import APIRouter, Cookie, Depends, Request, Response, status
from fastapi.responses import RedirectResponse
from yarl import URL

from ab_service.bff.dependencies import AppSettings, AuthClient, get_app_settings, get_auth_client
from ab_service.bff.models import Environment

logger = logging.getLogger(__name__)
router = APIRouter(prefix="/auth", tags=["Auth"])


@router.get("/login")
async def login(
    auth_client: Annotated[AuthClient, Depends(get_auth_client)],
    request: Request,
    return_to: str | None = None,
):
    """Return the current user's identity context."""
    # default return_to to the base URL of wherever the request came from (e.g. FE app)
    if not return_to:
        return_to = URL(str(request.base_url)).with_path("/").human_repr()

    # valid return to URL must have a hostname (to determine cookie domain for later);
    parsed_url = URL(return_to)
    hostname = parsed_url.host
    if not hostname:
        raise ValueError("Invalid return_to URL: hostname could not be determined")

    # kickstart the oauth 2.0 login flow by getting the authorization URL to redirect the user to
    login_payload = await auth_client.get_login_url_login_post(
        data=LoginRequest(
            scope="openid email profile offline_access entitlements",
            app_context={
                # after the token exchange, to a final return back to the original
                # requested page.
                "return_to": return_to,
            }
        )
    )
    return RedirectResponse(url=login_payload.url, status_code=status.HTTP_302_FOUND)


@router.get("/refresh", response_model=OAuth2TokenExposed)
async def refresh(
    auth_client: Annotated[AuthClient, Depends(get_auth_client)],
    request: Request,
    response: Response,
    app_settings: Annotated[AppSettings, Depends(get_app_settings)],
    refresh_token: Annotated[str | None, Cookie(alias="refresh_token")] = None,
    return_to: str | None = None,
):
    """Return the current user's identity context."""
    # default return_to to the base URL of wherever the request came from (e.g. FE app)
    if not return_to:
        return_to = URL(str(request.base_url)).with_path("/").human_repr()

    login_redirect = URL("/auth/login").with_query({"return_to": return_to}).human_repr()
    if not refresh_token:
        logger.debug("Missing refresh token, redirecting to login")
        return RedirectResponse(url=login_redirect, status_code=status.HTTP_302_FOUND)

    # exchange the auth code for tokens
    try:
        token = await auth_client.refresh_token_refresh_post(
            data=RefreshTokenRequest(
                refresh_token=refresh_token,
            ),
        )
    except Exception as e:
        logger.debug("Token refresh failed, redirecting to login", exc_info=e)
        return RedirectResponse(url=login_redirect, status_code=status.HTTP_302_FOUND)

    return _handle_cookie_redirect(return_to, token=token, app_settings=app_settings)


@router.get("/callback", response_model=OAuth2TokenExposed)
async def callback(
    auth_client: Annotated[AuthClient, Depends(get_auth_client)],
    app_settings: Annotated[AppSettings, Depends(get_app_settings)],
    request: Request,
):
    """Return the current user's identity context."""
    # exchange the auth code for tokens
    token = await auth_client.callback_callback_get(
        redirect_url=str(request.url),
    )

    # determine return_to url for redirect after login and cookie domain for setting cookies
    app_context = token.app_context or {}
    return_to = app_context.get("return_to", "/")

    return _handle_cookie_redirect(return_to, token=token, app_settings=app_settings)


@router.get("/logout")
async def logout(
    request: Request,
    app_settings: Annotated[AppSettings, Depends(get_app_settings)],
    return_to: str | None = None,
):
    """Logout the user by clearing cookies and redirecting to the given return_to URL (or base URL if not provided)."""
    # Where to send the browser after logout
    if not return_to:
        return_to = URL(str(request.base_url)).with_path("/").human_repr()

    # no token forces cookie clearing in the redirect response
    return _handle_cookie_redirect(return_to, token=None, app_settings=app_settings)


@router.get("/me", response_model=IdentityContext)
async def me(
    request: Request,
    identity_context: Annotated[IdentityContext, Depends(get_identity_context)],
):
    """Return the current user's identity context."""
    return identity_context


def _get_cookie_domain(url: str) -> str | None:
    parsed_url = URL(url)
    hostname = parsed_url.host
    cookie_domain = None

    # localhost / ip addresses should not have a cookie domain,
    # but other hostnames should be set with a cookie domain
    # of the base domain (e.g. "example.com" or ".example.com")
    # to allow subdomains to access the cookies
    if hostname and hostname != "localhost":
        try:
            ipaddress.ip_address(hostname)
        except ValueError:
            cookie_domain = f".{hostname}"

    return cookie_domain


def _handle_cookie_redirect(
    return_to: str,
    *,
    token: OAuth2TokenExposed | None = None,
    app_settings: AppSettings,
) -> RedirectResponse:
    """Helper to prepare the redirect response with cookies set for the given token and return_to URL."""
    # determine cookie domain (based off where the user should be redirected back to after login, which is typically the FE app)
    cookie_domain = _get_cookie_domain(return_to)

    # prepare redirect response to frontend with cookies set
    redirect = RedirectResponse(
        url=return_to,
        status_code=status.HTTP_302_FOUND,
    )

    # set frontend cookies with api tokens
    cookie_kwargs = dict(
        httponly=True,
        domain=cookie_domain,
        secure=app_settings.APP_ENV in {Environment.PRODUCTION, Environment.STAGING},
        samesite="lax",
        path="/",
    )

    # save token to cookies
    if token:
        for key in ("access_token", "refresh_token", "id_token"):
            value = getattr(token, key)
            if value:
                redirect.set_cookie(key, value, **cookie_kwargs)
            else:
                logger.warning(f"No {key} returned from callback; frontend may not work as expected")

    # clear token cookies
    else:
        for key in ("access_token", "refresh_token", "id_token"):
            redirect.delete_cookie(key, **cookie_kwargs)

    return redirect



"""Dependency providers for the BFF service.

This module exposes dependency-injected factories for the database and
remote service clients used by the BFF routes.
"""

from typing import Annotated

from ab_client.auth_client import AsyncClient as AuthClient
from ab_client.token_issuer import AsyncClient as TokenIssuerClient
from ab_client.token_issuer_store import AsyncClient as TokenIssuerStoreClient
from ab_client.token_store import AsyncClient as TokenStoreClient
from ab_client.token_validator import AsyncClient as TokenValidatorClient
from ab_client.user import AsyncClient as UserClient
from ab_core.database.databases import Database
from ab_core.dependency import Depends, inject, sentinel
from ab_core.dependency.loaders.environment_object import ObjectLoaderEnvironment

from .settings import AppSettings


@inject
async def get_app_settings(
    *,
    _app_settings: Annotated[
        AppSettings,
        Depends(
            ObjectLoaderEnvironment[AppSettings](env_prefix=""),
            persist=True,
        ),
    ] = sentinel(),
):
    yield _app_settings


@inject
async def get_database(
    *,
    _db: Annotated[
        Database,
        Depends(
            Database,
            persist=True,
        ),
    ] = sentinel(),
):
    yield _db


@inject
async def get_auth_client(
    *,
    _auth_client: Annotated[
        AuthClient,
        Depends(
            ObjectLoaderEnvironment[AuthClient](env_prefix="AUTH_SERVICE"),
            persist=True,
        ),
    ] = sentinel(),
):
    yield _auth_client


@inject
async def get_token_validator_client(
    *,
    _token_validator_client: Annotated[
        TokenValidatorClient,
        Depends(
            ObjectLoaderEnvironment[TokenValidatorClient](env_prefix="TOKEN_VALIDATOR_SERVICE"),
            persist=True,
        ),
    ] = sentinel(),
):
    yield _token_validator_client


@inject
async def get_user_client(
    *,
    _user_client: Annotated[
        UserClient,
        Depends(
            ObjectLoaderEnvironment[UserClient](env_prefix="USER_SERVICE"),
            persist=True,
        ),
    ] = sentinel(),
):
    yield _user_client


@inject
async def get_token_issuer_client(
    *,
    _token_issuer_client: Annotated[
        TokenIssuerClient,
        Depends(
            ObjectLoaderEnvironment[TokenIssuerClient](env_prefix="TOKEN_ISSUER_SERVICE"),
            persist=True,
        ),
    ] = sentinel(),
):
    yield _token_issuer_client


@inject
async def get_token_issuer_store_client(
    *,
    _token_issuer_store_client: Annotated[
        TokenIssuerStoreClient,
        Depends(
            ObjectLoaderEnvironment[TokenIssuerStoreClient](env_prefix="TOKEN_ISSUER_STORE_SERVICE"),
            persist=True,
        ),
    ] = sentinel(),
):
    yield _token_issuer_store_client


@inject
async def get_token_store_client(
    *,
    _token_store_client: Annotated[
        TokenStoreClient,
        Depends(
            ObjectLoaderEnvironment[TokenStoreClient](env_prefix="TOKEN_STORE_SERVICE"),
            persist=True,
        ),
    ] = sentinel(),
):
    yield _token_store_client


__all__ = [
    get_app_settings,
    AppSettings,
    get_database,
    Database,
    get_auth_client,
    AuthClient,
    get_token_validator_client,
    TokenValidatorClient,
    get_user_client,
    UserClient,
    get_token_issuer_client,
    TokenIssuerClient,
    get_token_issuer_store_client,
    TokenIssuerStoreClient,
    get_token_store_client,
    TokenStoreClient,
]



"""Main application for the User Service."""

from contextlib import asynccontextmanager
from typing import Annotated

from ab_core.alembic_auto_migrate.service import AlembicAutoMigrate
from ab_core.dependency import Depends, Load, inject
from ab_core.dependency.loaders import ObjectLoaderEnvironment
from ab_core.logging.config import LoggingConfig
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette.middleware.trustedhost import TrustedHostMiddleware
from uvicorn.middleware.proxy_headers import ProxyHeadersMiddleware

from ab_service.bff.dependencies import (
    AppSettings,
    AuthClient,
    Database,
    TokenIssuerClient,
    TokenIssuerStoreClient,
    TokenStoreClient,
    TokenValidatorClient,
    UserClient,
    get_auth_client,
    get_database,
    get_token_issuer_client,
    get_token_issuer_store_client,
    get_token_store_client,
    get_token_validator_client,
    get_user_client,
)
from ab_service.bff.routes.auth import router as auth_router


@asynccontextmanager
@inject
async def lifespan(
    _app: FastAPI,
    _db: Annotated[
        Database,
        Depends(
            get_database,
            persist=True,
        ),
    ],  # cold start load db into cache
    logging_config: Annotated[
        LoggingConfig,
        Depends(
            LoggingConfig,
            persist=True,
        ),
    ],
    alembic_auto_migrate: Annotated[
        AlembicAutoMigrate,
        Depends(
            AlembicAutoMigrate,
            persist=True,
        ),
    ],
    _auth_client: Annotated[
        AuthClient,
        Depends(
            get_auth_client,
            persist=True,
        ),
    ],
    _token_validator_client: Annotated[
        TokenValidatorClient,
        Depends(
            get_token_validator_client,
            persist=True,
        ),
    ],
    _user_client: Annotated[
        UserClient,
        Depends(
            get_user_client,
            persist=True,
        ),
    ],
    _token_issuer_client: Annotated[
        TokenIssuerClient,
        Depends(
            get_token_issuer_client,
            persist=True,
        ),
    ],
    _token_issuer_store_client: Annotated[
        TokenIssuerStoreClient,
        Depends(
            get_token_issuer_store_client,
            persist=True,
        ),
    ],
    _token_store_client: Annotated[
        TokenStoreClient,
        Depends(
            get_token_store_client,
            persist=True,
        ),
    ],
):
    """Lifespan context manager to handle startup and shutdown events."""
    logging_config.apply()
    alembic_auto_migrate.run()
    yield


app = FastAPI(lifespan=lifespan)
settings = Load(
    ObjectLoaderEnvironment[AppSettings](env_prefix=""),
    persist=True,
)
app.add_middleware(
    ProxyHeadersMiddleware,
    trusted_hosts=settings.TRUSTED_PROXY_IPS,
)
app.add_middleware(
    TrustedHostMiddleware,
    allowed_hosts=settings.ALLOWED_HOSTS,
)
app.add_middleware(
    CORSMiddleware,
    allow_origins=settings.ALLOWED_ORIGINS,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(auth_router)
# app.include_router(token_issuer_router)



"""User-related API routes."""

from typing import Annotated

from ab_client.token_issuer_store import (
    CreateTokenIssuerRequest as TokenIssuerStoreCreateTokenIssuerRequest,
)
from ab_client.token_issuer_store import (
    ManagedTokenIssuer,
)
from ab_core.identity_context.dependency import IdentityContext, get_identity_context
from fastapi import APIRouter, Depends

from ab_service.bff.dependencies import TokenIssuerStoreClient, get_token_issuer_store_client
from ab_service.bff.schema.token_issuer import CreateTokenIssuerRequest

router = APIRouter(prefix="/token_issuer", tags=["Token Issuer"])


@router.get("/{token_issuer_id}", response_model=ManagedTokenIssuer)
async def get_token_issuer(
    token_issuer_id: str,
    _identity_context: Annotated[IdentityContext, Depends(get_identity_context)],
    token_issuer_store_client: Annotated[TokenIssuerStoreClient, Depends(get_token_issuer_store_client)],
):
    """Lookup a token issuer by its ID."""
    return await token_issuer_store_client.get_token_issuer__id__get(
        id=token_issuer_id,
        # TODO: lookup should include the user id, ensuring
        #       users can only access their own token issuers
    )


@router.get("", response_model=ManagedTokenIssuer)
async def create_token_issuer(
    request: CreateTokenIssuerRequest,
    identity_context: Annotated[IdentityContext, Depends(get_identity_context)],
    token_issuer_store_client: Annotated[TokenIssuerStoreClient, Depends(get_token_issuer_store_client)],
):
    """Create a new token issuer."""
    return await token_issuer_store_client.create_token_issuer_post(
        data=TokenIssuerStoreCreateTokenIssuerRequest(
            name=request.name,
            token_issuer=request.token_issuer,
            created_by=identity_context.user.id,
        ),
    )



from packaging.markers import Environment
from pydantic import Field
from pydantic_settings import BaseSettings

from .models import Environment


class AppSettings(BaseSettings):
    """BFF Settings"""

    APP_ENV: Environment = Field(
        default=Environment.PRODUCTION,
        description="The application environment. Used for conditional logic and debugging. Set to 'development' for local dev.",
    )
    TRUSTED_PROXY_IPS: list[str] = Field(
        default_factory=list,
        description="List of trusted proxy IPs for ProxyHeadersMiddleware. Should be set to the IPs of proxies that may be forwarding requests to the BFF (e.g. API Gateway).",
    )
    ALLOWED_HOSTS: list[str] = Field(
        default_factory=list,
        description="List of allowed hosts for TrustedHostMiddleware. Should be set to the BE app hostname in production.",
    )
    ALLOWED_ORIGINS: list[str] = Field(
        default_factory=list,
        description="List of allowed originals for CORSMiddleware. Should be set to the FE app URL in production.",
    )


"""Pydantic schemas for token issuer endpoints.

Defines request/response models used by the token issuer API.
"""

from ab_client.token_issuer_store import TokenIssuer
from pydantic import BaseModel


class CreateTokenIssuerRequest(BaseModel):
    """Request model for creating a token issuer.

    Attributes:
        name: str: Human-friendly name for the issuer.
        token_issuer: TokenIssuer: The issuer configuration object.

    """

    name: str
    token_issuer: TokenIssuer
